@@ -1,67 +1,38 @@
 #!/usr/bin/env python3
-"""pyllelic: a tool for detection of allelic-specific varation in DNA sequencing.
-
-## Example exploratory use in jupyter notebook:
-
-    import pyllelic
-
-    # Setup directories and promoter of interest
-    pyllelic.set_up_env_variables(
-        base_path="/Users/abonham/documents/test_allelic/",
-        prom_file="TERT-promoter-genomic-sequence.txt",
-        prom_start="1293000",
-        prom_end="1296000",
-        chrom="5",
-        offset=1298163,
-    )
-
-    # find bam files
-    files_set = pyllelic.make_list_of_bam_files()
-
-    # index and create bam_output folders/files
-    positions = pyllelic.index_and_fetch(files_set)
-
-    # write out genome strings in bam_output folders
-    pyllelic.genome_parsing()
-
-    # pull out the cell types available for analysis
-    cell_types = extract_cell_types(files_set)
-
-    # run quma, get dfs
-    df_list = run_quma_and_compile_list_of_df(cell_types, filename)
-
-    # process means data from dataframes
-    means_df = process_means(df_list, positions, files_set)
-
-    # process modes data from dataframes
-    modes_df = process_modes(df_list, positions, cell_types)
-
-    # find difference between mean and mode
-    diff_df = find_diffs(means_df, modes_df)
-
-    # write output to excel files
-    write_means_modes_diffs(means_df, modes_df, diffs_df, filename)
+"""pyllelic: a tool for detection of allelic-specific variation
+   in reduced representation bisulfate DNA sequencing.
 """
 
-# Imports
-import pandas as pd
-import numpy as np
-import pysam
 import os
+import signal
 import sys
-from skbio.alignment import StripedSmithWaterman
-import plotly.graph_objects as go
-import subprocess
+from multiprocessing import Pool, cpu_count
 from pathlib import Path
+from typing import Any, Dict, List, Optional, Set, Tuple, Union
+
+import numpy as np
+import pandas as pd
+import plotly.graph_objects as go
+import pysam
+from Bio import pairwise2
 from scipy import stats
+
+# from skbio.alignment import StripedSmithWaterman
 from tqdm.notebook import tqdm
-from typing import List, Dict, Set, Optional, Tuple, Any, Union
-from .config import Config
+
 from . import quma
+from .config import Config
+
+import logging
+
+logging.basicConfig(filename="pyllelic_test.log", level=logging.DEBUG)
 
 # Initialize shared configuration object
 config = Config()
 
+# Initialized multiprocessing limits
+NUM_THREADS = cpu_count() - 1
+
 
 def set_up_env_variables(
     base_path: str,
@@ -103,7 +74,7 @@ def main() -> None:
 
     if sys.argv[1]:
         filename: str = sys.argv[1]
-    else:
+    else:  # pragma: no cover
         filename = "output.xlsx"
 
     files_set: List[str] = make_list_of_bam_files()
@@ -193,7 +164,7 @@ def run_sam_and_extract_df(sams: Path, process: bool = True) -> pd.Index:
 
     # Index samfile if index file doesn't exist
     if not sams.with_suffix(".bai").exists():
-        _: str = samtools_index(sams)  # we don't care what the output is
+        _: bool = pysam_index(sams)  # we don't care what the output is
 
     # Grab the promoter region of interest
     samm: pysam.AlignmentFile = pysam.AlignmentFile(sams, "rb")
@@ -243,20 +214,38 @@ def write_bam_output_files(sams: Path, positions: List[str], df: pd.DataFrame) -
     for each1 in positions:
         alignments: List = []
 
+        # # Set up query using alignment algorithm
+        # query_sequence: List[str] = df.loc[each1].head(1).tolist()[0]
+        # query: StripedSmithWaterman = StripedSmithWaterman(query_sequence)
+
+        # # Set up sequences to check for alignment
+        # target_sequences: List[str] = df.loc[each1].tolist()
+        # for target_sequence in target_sequences:
+        #     alignment = query(target_sequence)
+        #     alignments.append(alignment)
+
+        # read_file: List[str] = []
+        # for index, each in enumerate(alignments):
+        #     read_file.append(str(">read" + str(index)))
+        #     read_file.append(each.aligned_target_sequence)
+        #     # returns aligned target sequence
+
         # Set up query using alignment algorithm
         query_sequence: List[str] = df.loc[each1].head(1).tolist()[0]
-        query: StripedSmithWaterman = StripedSmithWaterman(query_sequence)
 
         # Set up sequences to check for alignment
         target_sequences: List[str] = df.loc[each1].tolist()
         for target_sequence in target_sequences:
-            alignment = query(target_sequence)
-            alignments.append(alignment)
+            alignment = pairwise2.align.localms(
+                query_sequence, target_sequence, 2, -3, -5, -2
+            )
+            aligned_segment = alignment[0].seqA[alignment[0].start : alignment[0].end]
+            alignments.append(aligned_segment)
 
         read_file: List[str] = []
         for index, each in enumerate(alignments):
             read_file.append(str(">read" + str(index)))
-            read_file.append(each.aligned_target_sequence)
+            read_file.append(each[0].seqB)
             # returns aligned target sequence
 
         write_individual_bam_file(sam_name, each1, read_file)
@@ -278,24 +267,19 @@ def write_individual_bam_file(
             file_handler.write("{}\n".format(item))
 
 
-def samtools_index(sams: Path) -> str:
-    """Helper function to run external samtools index tool.
+def pysam_index(bamfile: Path) -> bool:
+    """Helper function to run external samtools index.
 
     Args:
-        sams (Path): filepath to samfile
+        bamfile (Path): filepath to bam file
 
     Returns:
-        str: output from samtools index shell command, usually discarded
+        bool: verification of samtools command, usually discarded
     """
 
-    command: List[str] = ["samtools", "index", os.fspath(sams)]
+    pysam.index(os.fspath(bamfile))
 
-    output: subprocess.CompletedProcess = subprocess.run(
-        command, capture_output=True, text=True, check=True
-    )
-    out: str = output.stdout
-
-    return out
+    return True
 
 
 def genome_parsing(subfolders: List[Path] = None) -> None:
@@ -313,127 +297,77 @@ def genome_parsing(subfolders: List[Path] = None) -> None:
 
     # Wrap everything in processing them one at a time
     for folder in tqdm(subfolders, desc="Cell Lines: "):
-
-        # Grab list of read files in that directory:
-        raw_read_files: List[str] = os.listdir(folder)
-        read_files: List[str] = [
-            os.path.splitext(i)[0]
-            for i in raw_read_files
-            if not i.lstrip().startswith("g")
-            and not i.lstrip().startswith(".ip")
-            and not i.lstrip().startswith(".DS")  # mac .DS_store files
-        ]  # replace with regex?
-
-        # Now, process each file:
-        for read_name in read_files:
-            file_lines: List = []
-            # Grab the genomic sequence
-            file_lines.append(str(">genome" + str(read_name)))
-            file_lines.append(str(genome_range(read_name, genome_string)))
-
-            # Save the reads as a text file for each position
-            with open(folder.joinpath(f"g_{read_name}.txt"), "w") as file_handler:
-                for item in file_lines:
-                    file_handler.write(f"{item}\n")
+        _process_genome_parsing(folder, genome_string)
 
 
-def run_quma(directory: str, genomic_seq_file: str, reads_seq_file: str) -> str:
-    """Helper function to run external QUMA tool.
+def _process_genome_parsing(folder: Path, genome_string: str) -> None:
+    """Helper for processing genome file writing."""
+    # Grab list of read files in that directory:
+    raw_read_files: List[str] = os.listdir(folder)
+    read_files: List[str] = [
+        os.path.splitext(i)[0]
+        for i in raw_read_files
+        if not i.lstrip().startswith("g")
+        and not i.lstrip().startswith(".ip")
+        and not i.lstrip().startswith(".DS")  # mac .DS_store files
+    ]  # replace with regex?
 
-    Args:
-        directory (str): directory path to analyze
-        genomic_seq_file (str): text file with known genomic sequence
-        reads_seq_file (str): text file with experimental reads
+    # Now, process each file:
+    for read_name in read_files:
+        file_lines: List = []
+        # Grab the genomic sequence
+        file_lines.append(str(">genome" + str(read_name)))
+        file_lines.append(str(genome_range(read_name, genome_string)))
 
-    Returns:
-        bytes: shell output from quma command
-    """
-
-    quma_path: str = os.fspath(config.base_directory.joinpath("quma_cui"))
-    command: List[str] = [
-        "perl",
-        f"{quma_path}/quma.pl",
-        "-g",
-        f"{directory}/{genomic_seq_file}",
-        "-q",
-        f"{directory}/{reads_seq_file}",
-    ]
-
-    output: subprocess.CompletedProcess = subprocess.run(
-        command, text=True, capture_output=True, check=True
-    )
-    out: str = output.stdout
-    return out
+        # Save the reads as a text file for each position
+        with open(folder.joinpath(f"g_{read_name}.txt"), "w") as file_handler:
+            for item in file_lines:
+                file_handler.write(f"{item}\n")
 
 
-def access_quma(directory: str, genomic_seq_file: str, reads_seq_file: str) -> str:
+def access_quma(directory: Path, genomic_seq_file: str, reads_seq_file: str) -> str:
     """Helper function to run internal QUMA tool.
 
     Args:
-        directory (str): directory path to analyze
+        directory (Path): directory path to analyze
         genomic_seq_file (str): text file with known genomic sequence
         reads_seq_file (str): text file with experimental reads
 
     Returns:
         str: output from quma command
     """
-    return quma.quma_main(
+    result = quma.quma_main(
         f"{directory}/{genomic_seq_file}", f"{directory}/{reads_seq_file}"
     )
+    return result
 
 
-def quma_full(cell_types: List[str], filename: str) -> None:
-    """Run external QUMA methylation analysis on all specified cell lines,
-       writing out an excel file of results.
+def _init_worker():
+    """
+    Pool worker initializer for keyboard interrupt on Windows
+    """
+    signal.signal(signal.SIGINT, signal.SIG_IGN)
+
+
+def _thread_worker(folder: Path, read_name: str) -> pd.DataFrame:
+    """Queue worker for quma functions.
 
     Args:
-        cell_types (list[str]): list of cell lines in our dataset
-        filename (str): desired output filename for xlsx output
+        folder (Path): folder containing bam output
+        read_name (str): read position name
+
+    Returns:
+        pd.DataFrame: dataframe of quma results
     """
+    quma_result: str = access_quma(folder, f"g_{read_name}.txt", f"{read_name}.txt")
+    processed_quma: List[str] = process_raw_quma(quma_result)
+    # Next, add this readname to the holding data frame
+    int_df: pd.DataFrame = pd.DataFrame({read_name: processed_quma})
 
-    # Grab list of directories
-    subfolders: List[Path] = [p for p in config.bam_directory.iterdir() if p.is_dir()]
-
-    writer: pd.ExcelWriter = pd.ExcelWriter(config.base_directory.joinpath(filename))
-
-    # Wrap everything in processing them one at a time
-    for folder in tqdm(subfolders, desc="Cell Lines"):
-
-        # Get short name of cell_line
-        cell_line_name: str = folder.name.split("_")[1]
-
-        if set([cell_line_name]).intersection(set(cell_types)):
-            # Set up a holding data frame from all the data
-            holding_df: pd.DataFrame = pd.DataFrame()
-
-            # Grab list of read files in that directory:
-            raw_read_files: List[str] = os.listdir(folder)
-            read_files: List[str] = [
-                os.path.splitext(i)[0]
-                for i in raw_read_files
-                if i.endswith(".txt") and not i.lstrip().startswith("g")
-            ]
-
-            # Now, process each file:
-            for read_name in tqdm(read_files, desc="Positions", leave=False):
-
-                quma_result: str = access_quma(  # changed for internal quma
-                    folder, f"g_{read_name}.txt", f"{read_name}.txt"
-                )
-                processed_quma: List[str] = process_raw_quma(quma_result)
-                # Next, add this readname to the holding data frame
-                int_df: pd.DataFrame = pd.DataFrame({read_name: processed_quma})
-                holding_df = pd.concat([holding_df, int_df], axis=1)
-
-            # Now, save it to an excel file
-            holding_df.to_excel(writer, sheet_name=cell_line_name)
-
-            del holding_df
-
-    writer.save()
+    return int_df
 
 
-def quma_full_mp(cell_types: List[str], filename: str) -> None:
+def quma_full_threaded(cell_types: List[str], filename: str) -> None:
     """Run external QUMA methylation analysis on all specified cell lines,
        using multiprocessing library.
 
@@ -446,63 +380,69 @@ def quma_full_mp(cell_types: List[str], filename: str) -> None:
 
     writer: pd.ExcelWriter = pd.ExcelWriter(config.base_directory.joinpath(filename))
 
-    # Wrap everything in processing them one at a time
-    for folder in tqdm(subfolders, desc="Cell Lines"):
+    folders_to_use = [
+        folder
+        for folder in subfolders
+        if set([folder.name.split("_")[1]]).intersection(set(cell_types))
+    ]
 
-        # Get short name of cell_line
+    for folder in tqdm(folders_to_use, desc="Cell Lines"):
         cell_line_name: str = folder.name.split("_")[1]
 
-        if set([cell_line_name]).intersection(set(cell_types)):
-            # Set up a holding data frame from all the data
-            holding_df: pd.DataFrame = pd.DataFrame()
+        # Grab list of read files in that directory:
+        raw_read_files: List[str] = os.listdir(folder)
+        read_files: List[str] = [
+            os.path.splitext(i)[0]
+            for i in raw_read_files
+            if i.endswith(".txt") and not i.lstrip().startswith("g")
+        ]
 
-            # Grab list of read files in that directory:
-            raw_read_files: List[str] = os.listdir(folder)
-            read_files: List[str] = [
-                os.path.splitext(i)[0]
-                for i in raw_read_files
-                if i.endswith(".txt") and not i.lstrip().startswith("g")
-            ]
-            quma_path: str = os.fspath(config.base_directory.joinpath("quma_cui"))
-            command_list: List[List[str]] = [
-                [
-                    "python",
-                    f"{quma_path}/quma.py",
-                    "-g",
-                    f"{folder}/g_{read_pos}.txt",
-                    "-q",
-                    f"{folder}/{read_pos}.txt",
-                ]
-                for read_pos in read_files
-            ]
+        # Send for multiprocessing
+        holding_df = _pool_processing(read_files, folder)
+        # Now, save it to an excel file
+        holding_df.to_excel(writer, sheet_name=cell_line_name)
 
-            procs_dict = {
-                pos: subprocess.Popen(
-                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
-                )
-                for pos, cmd in zip(read_files, command_list)
-            }
-
-            raw_results = {}
-            for name, proc in procs_dict.items():
-                r = proc.communicate()[0].decode("utf-8")
-                raw_results[name] = r
-
-            for key, result in raw_results.items():
-                processed_quma: List[str] = process_raw_quma(result)
-
-                # Next, add this readname to the holding data frame
-                int_df: pd.DataFrame = pd.DataFrame({key: processed_quma})
-                holding_df = pd.concat([holding_df, int_df], axis=1)
-
-            # Now, save it to an excel file
-            holding_df.to_excel(writer, sheet_name=cell_line_name)
-
-            del holding_df
+        del holding_df
 
     writer.save()
 
 
+def _pool_processing(read_files: List[str], folder: Path) -> pd.DataFrame:
+    """Helper to run quma in a multiprocessing pool.
+
+    Args:
+        read_files (List[str]): list of files to process
+        folder (Path): path to bam_output folder
+
+    Returns:
+        pd.DataFrame: dataframe of quma results
+    """
+
+    # Set up a holding data frame from all the data
+    holding_df: pd.DataFrame = pd.DataFrame()
+
+    returns: List[Any] = []
+    with Pool(NUM_THREADS, _init_worker) as pool:
+        for read_name in read_files:
+
+            result = pool.apply_async(
+                _thread_worker,
+                (folder, read_name),
+            )
+            returns.append(result)
+
+        results = [result.get() for result in returns]
+
+    # Add to excel file
+    for each in results:
+        try:
+            holding_df = pd.concat([holding_df, each], axis=1)
+        except (AttributeError, KeyError):
+            pass
+
+    return holding_df
+
+
 def process_raw_quma(quma_result: str) -> List[str]:
     """Spit and process raw quma results into a pandas dataframe.
 
@@ -517,10 +457,10 @@ def process_raw_quma(quma_result: str) -> List[str]:
     for line in quma_result.splitlines():
         if not line.lstrip().startswith("g"):
             fields: List[str] = line.split("\t")
-            # if float(fields[7]) < 70:  # CONSIDER: checks for high mercent match
-            #     dots.append("FAIL")
-            # else:
-            dots.append(fields[13])
+            if float(fields[7]) < 70:  # checks for high mercent match
+                dots.append("FAIL")
+            else:
+                dots.append(fields[13])
     # Sort dots output by number of "1"
     return sorted(dots, key=lambda t: t.count("1"))
 
@@ -546,7 +486,7 @@ def run_quma_and_compile_list_of_df(
     """
 
     if run_quma:
-        quma_full(cell_types, filename)  # Changed for multiprocessing of quma
+        quma_full_threaded(cell_types, filename)  # Changed for multiprocessing of quma
 
     dict_of_df: Dict[str, pd.DataFrame] = read_df_of_quma_results(filename)
 
@@ -590,7 +530,7 @@ def process_means(
     working_df: pd.DataFrame = pd.DataFrame()
     for pos in positions:
         working_df[pos] = ""
-        for key, each in dict_of_dfs.items():
+        for key in dict_of_dfs.keys():
             values_list: List[float] = return_read_values(pos, key, dict_of_dfs)
 
             if values_list:
@@ -623,7 +563,7 @@ def process_modes(
     working_df: pd.DataFrame = pd.DataFrame()
     for pos in positions:
         working_df[pos] = ""
-        for key, each in dict_of_dfs.items():
+        for key in dict_of_dfs.keys():
             values_list: List[float] = return_read_values(pos, key, dict_of_dfs)
 
             if values_list:
@@ -655,7 +595,7 @@ def return_individual_data(
     working_df: pd.DataFrame = pd.DataFrame()
     for pos in tqdm(positions, desc="Position"):
         working_df[pos] = ""  # Create position column in dataframe
-        for key, each in tqdm(dict_of_dfs.items(), desc="Cell Line", leave=False):
+        for key in tqdm(dict_of_dfs.keys(), desc="Cell Line", leave=False):
             values_list: List[float] = return_read_values(pos, key, dict_of_dfs)
             if values_list:
                 data_for_df: Union[List[float], float] = values_list
@@ -671,8 +611,8 @@ def return_read_values(
     pos: str,
     key: str,
     dict_of_dfs: Dict[str, pd.DataFrame],
-    min_reads: int = 4,
-    min_sites: int = 2,
+    min_reads: int = 1,
+    min_sites: int = 1,
 ) -> List[float]:
     """Given a genomic position, cell line, and data, find fractional methylation
        values for each read in the dataset.
@@ -845,8 +785,7 @@ def anderson_darling_test(
     if np.all(pd.notnull(raw_list)):
         stat: float
         crits: List[Any]
-        sigs: List[Any]
-        stat, crits, sigs = stats.anderson(raw_list)
+        stat, crits, _ = stats.anderson(raw_list)
         if stat > crits[4]:
             is_sig = True
         else:
