@@ -1,16 +1,22 @@
+#!/usr/bin/env python3
+"""Tools to quantify methylation in reduced representation
+   bisulfite sequencing reads."""
+
 # Adapted from QUMA CLI: http://quma.cdb.riken.jp/
 # Licensed under GPLv3
 # perl conversion by: https://freelancer.com/u/Zubayerskd
 
-import time
-from datetime import datetime
 import re
-import os
 from io import StringIO
-from typing import List, Tuple, Dict, Any
+from typing import Any, Dict, List, Tuple, Optional
+
 from Bio import pairwise2
 from Bio.Align import substitution_matrices
 
+import logging
+
+logging.basicConfig(filename="quma_test.log", level=logging.DEBUG)
+
 MAX_LINE_LENGTH = 60
 MAT = """    A   T   G   C   S   W   R   Y   K   M   B   V   H   D   N   U
     A   5  -4  -4  -4  -4   1   1  -4  -4   1  -4  -1  -1  -1  -2  -4
@@ -50,21 +56,6 @@ def check_char_in_allowed(seq: str, pattern: str) -> str:
     return new
 
 
-def make_time(given_time: float = None) -> str:
-    """Convert floating time into string formatted timestamp.
-
-    Args:
-        given_time (float, optional): floating point time. Defaults to None.
-
-    Returns:
-        str: String-formatted timestamp.
-    """
-    curr_time: float = given_time or time.time()
-
-    dt: datetime = datetime.fromtimestamp(curr_time)
-    return dt.strftime("%Y%m%d%H%M%S")
-
-
 def curate_seq(seq: str) -> str:
     """Curate a sequence to only have allowed characters.
 
@@ -86,7 +77,7 @@ def parse_seq(seq: str) -> str:
     Returns:
         str: sequence string
     """
-    _ = ""  # was com
+    _ = ""
     seq = re.sub(r"\r\n", "\n", seq)
     seq = re.sub(r"\r", "\n", seq)
     seq = seq.upper()
@@ -95,19 +86,25 @@ def parse_seq(seq: str) -> str:
     if re.findall(reg, seq):
         seq = re.sub(r"^\s*>(.*)?\n", "", seq)
 
-    elif re.findall(r"^ORIGIN\s*\n((\s+(\d+(\s+\w+)+))+)\s*\n//", seq, re.MULTILINE):
+    elif re.findall(
+        r"^ORIGIN\s*\n((\s+(\d+(\s+\w+)+))+)\s*\n//", seq, re.MULTILINE
+    ):  # pragma: no cover
         seq = re.findall(reg, seq, re.MULTILINE)[0]
 
-    elif re.findall(r"^SQ\s+SEQUENCE.*\n((\s+(\w+\s+)+\d+)+)\n\/\/", seq, re.MULTILINE):
+    elif re.findall(
+        r"^SQ\s+SEQUENCE.*\n((\s+(\w+\s+)+\d+)+)\n\/\/", seq, re.MULTILINE
+    ):  # pragma: no cover
         seq = re.findall(
             r"^SQ\s+SEQUENCE.*\n((\s+(\w+\s+)+\d+)+)\n\/\/", seq, re.MULTILINE
         )[0]
 
-    elif re.findall(r"\.\.\s*\n((\s+(\d+(\s+\w+)+))+)\s*", seq, re.MULTILINE):
+    elif re.findall(
+        r"\.\.\s*\n((\s+(\d+(\s+\w+)+))+)\s*", seq, re.MULTILINE
+    ):  # pragma: no cover
         seq = re.findall(r"\.\.\s*\n((\s+(\d+(\s+\w+)+))+)\s*", seq, re.MULTILINE)[0]
-    elif re.findall(r"^\s*>.+\s.+", seq, re.MULTILINE):
+    elif re.findall(r"^\s*>.+\s.+", seq, re.MULTILINE):  # pragma: no cover
         seq = re.findall(r"^\s*>(.+?)\s(?=.+)", seq, re.MULTILINE)[0]
-        _ = seq  # was com
+        _ = seq
 
     return curate_seq(seq)
 
@@ -151,7 +148,7 @@ def multi_fasta_parse(multi: Any) -> List[Dict[str, str]]:
     for line in multi:
 
         if ">" in line:
-            if fa and not fa["seq"]:
+            if fa and not fa["seq"]:  # pragma: no cover
                 biseq.pop()
             fa = {"com": line}
             fa["com"] = re.sub(r"^>", "", fa["com"])
@@ -161,7 +158,7 @@ def multi_fasta_parse(multi: Any) -> List[Dict[str, str]]:
             line = curate_seq(line)
             if line == "":
                 continue
-            if not fa:
+            if not fa:  # pragma: no cover
                 return None  # does this ever happen?
             try:
                 fa["seq"] += line.upper()
@@ -170,8 +167,8 @@ def multi_fasta_parse(multi: Any) -> List[Dict[str, str]]:
 
     if fa:
         try:
-            _ = fa["seq"]  # was temp
-        except KeyError:
+            _ = fa["seq"]
+        except KeyError:  # pragma: no cover
             biseq.pop()
     return biseq
 
@@ -252,7 +249,7 @@ def fasta_print(
         line (int, optional): max length to process. Defaults to None.
         add (bool, optional): append to file flag. Defaults to None.
     """
-    if add:
+    if add:  # pragma: no cover
         with open(path, "a") as f:
             f.write(fasta_make(seq, seq_name, line))
 
@@ -364,8 +361,11 @@ def align_seq_and_generate_stats(
     bio_gseq = gfile.splitlines()[1]
     bio_qseq = qfile.splitlines()[1]
     # flipping to correct output comparison
-    bio_alignments = pairwise2.align.globalds(bio_qseq, bio_gseq, MATRIX, -10, -0.5)
-    fh_ = f">genome\n{bio_alignments[0][0]}\n>que\n{bio_alignments[0][1]}\n"
+    bio_alignments = pairwise2.align.localds(bio_qseq, bio_gseq, MATRIX, -10, -0.5)
+    genome_ali = bio_alignments[0].seqB[bio_alignments[0].start : bio_alignments[0].end]
+    query_ali = bio_alignments[0].seqA[bio_alignments[0].start : bio_alignments[0].end]
+    fh_ = f">genome\n{genome_ali}\n>que\n{query_ali}\n"
+    # logging.debug(f"gseq={bio_gseq},aligned={genome_ali}")
     fh: List[str] = fh_.split("\n")
 
     for i, line in enumerate(fh):
@@ -377,7 +377,7 @@ def align_seq_and_generate_stats(
     ref["qAli"] = ref["qAli"].replace(" ", "-")
     ref["gAli"] = ref["gAli"].replace(" ", "-")
 
-    if len(ref["qAli"]) != len(ref["gAli"]):
+    if len(ref["qAli"]) != len(ref["gAli"]):  # pragma: no cover
         print("qAli len != gAli len")
         # sys.exit()
 
@@ -385,6 +385,31 @@ def align_seq_and_generate_stats(
     return results
 
 
+def _generate_summary_stats(ref: Dict[str, Any]) -> Dict[str, Any]:
+    """Helper to generate summary statistics in results dictionary."""
+
+    if ref["conv"] + ref["unconv"] != 0:
+        ref["pconv"] = _percentage(ref["conv"], ref["unconv"], calc_type="sum")
+    else:
+        ref["pconv"] = 0
+
+    ref["perc"] = _percentage(ref["match"], ref["aliLen"], calc_type="total")
+    ref["perc"] = float(ref["perc"])
+    ref["pconv"] = float(ref["pconv"])
+    ref["aliMis"] = ref["aliLen"] - ref["match"]
+
+    return ref
+
+
+def _percentage(a: int, b: int, calc_type: str) -> Optional[str]:
+    """Helper to return percentages."""
+    if calc_type == "sum":
+        return f"{(100 * a / (a + b)):3.1f}"
+    if calc_type == "total":
+        return f"{(100 * a / b):3.1f}"
+    return None  # pragma: no cover
+
+
 def process_alignment_matches(
     ref: Dict[str, Any], cpg: Dict[str, Any]
 ) -> Dict[str, Any]:
@@ -397,34 +422,23 @@ def process_alignment_matches(
     Returns:
         Dict[str, Any]: results dictionary
     """
-    # ref = ref
-    # cpg = cpg
-    gAli = qAli = ""
-    # fl = 0
-
-    for i, val in enumerate(ref["gAli"]):
-        # removed gap stripping, not needed for alignment function
-        # if val == "-" and fl == 0:
-        #     continue
-        # fl = 1
-        gAli += val
-        qAli += ref["qAli"][i]
-
-    # gAli = gAli.rstrip("-")
-    qAli = qAli[: len(gAli)]
+    gAli = ref["gAli"]
+    qAli = ref["qAli"][: len(gAli)]
 
+    _, cpg, _ = find_cpg(gAli)
     ref["aliLen"] = len(qAli)
 
+    # Loop through sequence looking for CpG conversion
     j = 0
     for i in range(ref["aliLen"]):
 
         g = gAli[i]
         q = qAli[i]
 
-        # if g != "-":
-        #     j += 1
-        j += 1
-        if q == g or (g == "C" and q == "T"):
+        if g != "-":
+            j += 1
+
+        if q == g or g == "C" and q == "T":
             ref["match"] += 1
 
         if g == "-":
@@ -433,49 +447,45 @@ def process_alignment_matches(
 
         if q == "-":
             ref["gap"] += 1
-            try:
-                _ = cpg[str(j - 1)]
+            if cpg.get(str(j - 1)):
                 ref["val"] += "-"
-            except KeyError:
-                pass
             continue
 
         if i == ref["aliLen"] - 1:
             break
 
+        # End if genome sequence isn't a C
         if g != "C":
             continue
 
-        try:
-            _ = cpg[str(j - 1)]
+        # If this is a CpG C on the genome, inspect query
+        if cpg.get(str(j - 1)):
             if q == "C":
                 ref["conv"] += 1
+                ref["menum"] += 1
+                ref["val"] += "1"
             elif q == "T":
                 ref["unconv"] += 1
-        except KeyError:
+                ref["val"] += "0"
+            else:
+                ref["val"] += q
             continue
 
-        if q == "C":
-            ref["menum"] += 1
-            ref["val"] += "1"
-        elif q == "T":
-            ref["val"] += "0"
-        else:
-            ref["val"] += q
+        # if q == "C":
+        #     ref["menum"] += 1
+        #     ref["val"] += "1"
+        # elif q == "T":
+        #     ref["val"] += "0"
+        # else:
+        #     ref["val"] += q
 
-    if ref["conv"] + ref["unconv"] != 0:
-        ref["pconv"] = "{:3.1f}".format(
-            100 * ref["conv"] / (ref["conv"] + ref["unconv"])
-        )
-    else:
-        ref["pconv"] = 0
+    results = _generate_summary_stats(ref)
 
-    ref["perc"] = "{:3.1f}".format(100 * ref["match"] / ref["aliLen"])
-    ref["perc"] = float(ref["perc"])
-    ref["pconv"] = float(ref["pconv"])
-
-    ref["aliMis"] = ref["aliLen"] - ref["match"]
-    return ref
+    # kludge:
+    if ref["val"] == "":
+        ref["val"] = "-"
+    # logging.debug(f"results={results}")
+    return results
 
 
 def process_fasta_output(
@@ -504,7 +514,7 @@ def process_fasta_output(
     pos = 0
     for fa in qseq:
         pos += 1
-        fa["pos"] = str(pos)  # was int
+        fa["pos"] = str(pos)
 
         qfilepF = fasta_output(fa["seq"], qfileF)
         qfilepR = fasta_output(rev_comp(fa["seq"]), qfileR)
@@ -543,10 +553,16 @@ def process_fasta_output(
             fdir = 1
 
         res = fres
-        dir = fdir
+        final_dir = fdir
         gdir = 1
 
-        ref: Dict[str, Any] = {"fa": fa, "res": res, "dir": dir, "gdir": gdir, "exc": 0}
+        ref: Dict[str, Any] = {
+            "fa": fa,
+            "res": res,
+            "dir": final_dir,
+            "gdir": gdir,
+            "exc": 0,
+        }
         if res["unconv"] > unc:
             ref["exc"] = 1
         if res["pconv"] > pcon:
@@ -624,18 +640,18 @@ def find_cpg(gseq: str) -> Tuple[List[str], Dict[str, Any], Dict[str, Any]]:
 def quma_main(gfile: str, qfile: str) -> str:
     """Run quma for quantification of methylation of bisulfite sequencing reads."""
 
-    t = make_time()
-    uid = "{}{:06d}".format(t, os.getpid())
+    # t = make_time()
+    # uid = "{}{:06d}".format(t, os.getpid())
 
     gseq = parse_genome(gfile)
     qseq = parse_biseq(qfile)
 
     positions, cpgf, cpgr = find_cpg(gseq)
 
-    qfileF = f"que{uid}F"
-    qfileR = f"que{uid}R"
-    gfileF = f"genome{uid}F"
-    gfileR = f"genome{uid}R"
+    qfileF = "queryF"  # f"que{uid}F"
+    qfileR = "queryR"  # f"que{uid}R"
+    gfileF = "genomeF"  # f"genome{uid}F"
+    gfileR = "genomeR"  # f"genome{uid}R"
 
     gfilepF = fasta_output(gseq, gfileF)
     gfilepR = fasta_output(gseq, gfileR)
